from dataclasses import dataclass, field
from datetime import datetime
from typing import Iterator, List, Optional, Union
from uuid import UUID, uuid4

from isar.apis.models.models import StartMissionResponse, TaskResponse
from isar.config.settings import settings
from isar.models.mission_metadata.mission_metadata import MissionMetadata
from robot_interface.models.mission import (
    InspectionStep,
    MotionStep,
    STEPS,
    Step,
    StepStatus,
)
from robot_interface.models.mission.step import DriveToPose
from .status import MissionStatus, TaskStatus


@dataclass
class Task:
    steps: List[STEPS]
    status: TaskStatus = field(default=TaskStatus.NotStarted, init=False)
    tag_id: Optional[str] = field(default=None)
    id: UUID = field(default_factory=uuid4, init=False)
    _iterator: Iterator = None

    def next_step(self) -> Step:
        step: Step = next(self._iterator)
        while step.status != StepStatus.NotStarted:
            step = next(self._iterator)
        return step

    def is_finished(self) -> bool:
        for step in self.steps:
            if step.status is StepStatus.Failed and isinstance(step, MotionStep):
                # One motion step has failed meaning the task as a whole should be
                # considered as failed
                return True

            elif (step.status is StepStatus.Failed) and isinstance(
                step, InspectionStep
            ):
                # It should be possible to perform several inspections per task. If
                # one out of many inspections fail the task is considered as
                # partially successful.
                continue

            elif step.status is StepStatus.Successful:
                # The task is complete once all steps are completed
                continue
            else:
                # Not all steps have been completed yet
                return False

        return True

    def update_task_status(self) -> None:
        for step in self.steps:
            if step.status is StepStatus.Failed and isinstance(step, MotionStep):
                self.status = TaskStatus.Failed
                return

            elif (step.status is StepStatus.Failed) and isinstance(
                step, InspectionStep
            ):
                self.status = TaskStatus.PartiallySuccessful
                continue

            elif step.status is StepStatus.Successful:
                continue

        if self.status is not TaskStatus.PartiallySuccessful:
            self.status = TaskStatus.Successful

        elif self._all_inspection_steps_failed():
            self.status = TaskStatus.Failed

    def reset_task(self):
        for step in self.steps:
            if isinstance(step, DriveToPose):
                step.status = StepStatus.NotStarted
            elif (
                isinstance(step, InspectionStep)
                and step.status == StepStatus.InProgress
            ):
                step.status = StepStatus.NotStarted
        self._iterator = iter(self.steps)

    def _all_inspection_steps_failed(self) -> bool:
        for step in self.steps:
            if isinstance(step, MotionStep):
                continue
            elif step.status is not StepStatus.Failed:
                return False

        return True

    def __post_init__(self) -> None:
        if self._iterator is None:
            self._iterator = iter(self.steps)

    def api_response(self) -> TaskResponse:
        return TaskResponse(
            id=self.id,
            tag_id=self.tag_id,
            steps=list(
                map(lambda x: {"id": x.id, "type": x.__class__.__name__}, self.steps)
            ),
        )


@dataclass
class Mission:
    tasks: List[Task]
    id: Union[UUID, int, str, None] = None
    status: MissionStatus = MissionStatus.NotStarted
    metadata: MissionMetadata = None

    def set_unique_id_and_metadata(self) -> None:
        self._set_unique_id()
        self.metadata = MissionMetadata(mission_id=self.id)

    def _set_unique_id(self) -> None:
        plant_short_name: str = settings.PLANT_SHORT_NAME
        robot_id: str = settings.ROBOT_ID
        now: datetime = datetime.utcnow()
        self.id = (
            f"{plant_short_name.upper()}{robot_id.upper()}"
            f"{now.strftime('%d%m%Y%H%M%S%f')[:-3]}"
        )

    def __post_init__(self) -> None:
        if self.id is None:
            self._set_unique_id()

        if self.metadata is None:
            self.metadata = MissionMetadata(mission_id=self.id)

    def api_response(self) -> StartMissionResponse:
        return StartMissionResponse(
            id=self.id,
            tasks=[task.api_response() for task in self.tasks],
        )



import json
import logging
import time
from datetime import datetime
from logging import Logger
from queue import Queue
from threading import Thread

from isar.config.settings import settings
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states_enum import States
from robot_interface.models.exceptions import RobotCommunicationException
from robot_interface.models.mission.status import RobotStatus
from robot_interface.robot_interface import RobotInterface
from robot_interface.telemetry.mqtt_client import MqttPublisher
from robot_interface.telemetry.payloads import RobotStatusPayload
from robot_interface.utilities.json_service import EnhancedJSONEncoder


class RobotStatusPublisher:
    def __init__(
        self, mqtt_queue: Queue, robot: RobotInterface, state_machine: StateMachine
    ):
        self.mqtt_publisher: MqttPublisher = MqttPublisher(mqtt_queue=mqtt_queue)
        self.robot: RobotInterface = robot
        self.state_machine: StateMachine = state_machine

    def _get_combined_robot_status(
        self, robot_status: RobotStatus, current_state: States
    ) -> RobotStatus:
        if robot_status == RobotStatus.Offline:
            return RobotStatus.Offline
        elif current_state == States.Idle and robot_status == RobotStatus.Available:
            return RobotStatus.Available
        elif current_state != States.Idle or robot_status == RobotStatus.Busy:
            return RobotStatus.Busy
        return None

    def run(self) -> None:
        robot_status_monitor: RobotStatusMonitor = RobotStatusMonitor(robot=self.robot)
        robot_status_thread: Thread = Thread(
            target=robot_status_monitor.run,
            name="Robot Status Monitor",
            daemon=True,
        )
        robot_status_thread.start()

        while True:
            combined_status: RobotStatus = self._get_combined_robot_status(
                robot_status=robot_status_monitor.robot_status,
                current_state=self.state_machine.current_state,
            )

            payload: RobotStatusPayload = RobotStatusPayload(
                robot_name=settings.ROBOT_ID,
                robot_status=combined_status,
                current_isar_state=self.state_machine.current_state,
                current_mission_id=self.state_machine.current_mission.id
                if self.state_machine.current_mission
                else None,
                current_task_id=self.state_machine.current_task.id
                if self.state_machine.current_task
                else None,
                current_step_id=self.state_machine.current_step.id
                if self.state_machine.current_step
                else None,
                timestamp=datetime.utcnow(),
            )

            self.mqtt_publisher.publish(
                topic=settings.TOPIC_ISAR_ROBOT_STATUS,
                payload=json.dumps(payload, cls=EnhancedJSONEncoder),
            )

            time.sleep(settings.ROBOT_STATUS_PUBLISH_INTERVAL)


class RobotStatusMonitor:
    def __init__(self, robot: RobotInterface):
        self.robot: RobotInterface = robot
        self.robot_status: RobotStatus = RobotStatus.Offline
        self.logger: Logger = logging.getLogger("robot_status_monitor")

    def run(self) -> None:
        while True:
            try:
                self.robot_status = self.robot.robot_status()
            except RobotCommunicationException:
                self.logger.warning(
                    "Failed to get robot status due to a communication exception"
                )
                time.sleep(settings.ROBOT_API_STATUS_POLL_INTERVAL)



import json
import logging

from azure.identity import DefaultAzureCredential
from injector import inject
from requests import HTTPError, RequestException
from requests_toolbelt import MultipartEncoder

from isar.config.settings import settings
from isar.models.mission_metadata.mission_metadata import MissionMetadata
from isar.services.auth.azure_credentials import AzureCredentials
from isar.services.service_connections.request_handler import RequestHandler
from isar.storage.storage_interface import StorageException, StorageInterface
from isar.storage.utilities import get_filename
from robot_interface.models.inspection.inspection import Inspection, ThermalVideo, Video


class SlimmStorage(StorageInterface):
    @inject
    def __init__(self, request_handler: RequestHandler) -> None:
        self.request_handler: RequestHandler = request_handler
        self.logger = logging.getLogger("uploader")

        self.credentials: DefaultAzureCredential = (
            AzureCredentials.get_azure_credentials()
        )

        client_id: str = settings.SLIMM_CLIENT_ID
        scope: str = settings.SLIMM_APP_SCOPE
        self.request_scope: str = f"{client_id}/{scope}"

        self.url: str = settings.SLIMM_API_URL

    def store(self, inspection: Inspection, metadata: MissionMetadata):
        filename: str = get_filename(
            mission_id=metadata.mission_id,
            inspection_type=type(inspection).__name__,
            inspection_id=inspection.id,
        )
        filename = f"{filename}.{inspection.metadata.file_type}"
        if type(inspection) in [Video, ThermalVideo]:
            self._store_video(filename, inspection, metadata)
        else:
            self._store_image(filename, inspection, metadata)

    def _store_image(
        self, filename: str, inspection: Inspection, metadata: MissionMetadata
    ):
        multiform_body: MultipartEncoder = self._construct_multiform_request_image(
            filename=filename, inspection=inspection, metadata=metadata
        )
        request_url: str = f"{self.url}/UploadSingleImage"
        self._ingest(
            inspection=inspection,
            multiform_body=multiform_body,
            request_url=request_url,
        )
        return

    def _store_video(
        self, filename: str, inspection: Inspection, metadata: MissionMetadata
    ):
        multiform_body: MultipartEncoder = self._construct_multiform_request_video(
            filename=filename, inspection=inspection, metadata=metadata
        )
        request_url = f"{self.url}/UploadSingleVideo"
        self._ingest(
            inspection=inspection,
            multiform_body=multiform_body,
            request_url=request_url,
        )
        return

    def _ingest(
        self, inspection: Inspection, multiform_body: MultipartEncoder, request_url: str
    ):
        token: str = self.credentials.get_token(self.request_scope).token
        try:
            response = self.request_handler.post(
                url=request_url,
                data=multiform_body,
                headers={
                    "Authorization": f"Bearer {token}",
                    "Content-Type": multiform_body.content_type,
                },
            )
            guid = json.loads(response.text)["guid"]
            self.logger.info(f"SLIMM upload GUID: {guid}")
        except (RequestException, HTTPError) as e:
            self.logger.warning(
                f"Failed to upload inspection: {inspection.id} to SLIMM due to a "
                f"request exception"
            )
            raise StorageException from e

    @staticmethod
    def _construct_multiform_request_image(
        filename: str, inspection: Inspection, metadata: MissionMetadata
    ):
        array_of_orientation = (
            inspection.metadata.pose.orientation.to_quat_array().tolist()
        )
        multiform_body: MultipartEncoder = MultipartEncoder(
            fields={
                "PlantFacilitySAPCode": metadata.plant_code,
                "InstCode": metadata.plant_short_name,
                "InternalClassification": metadata.data_classification,
                "IsoCountryCode": "NO",
                "Geodetic.CoordinateReferenceSystemCode": metadata.coordinate_reference_system,  # noqa: E501
                "Geodetic.VerticalCoordinateReferenceSystemCode": metadata.vertical_reference_system,  # noqa: E501
                "Geodetic.OrientationReferenceSystem": metadata.media_orientation_reference_system,  # noqa: E501
                "SensorCarrier.SensorCarrierId": metadata.robot_id,
                "SensorCarrier.ModelName": metadata.robot_model,
                "Mission.MissionId": metadata.mission_id,
                "Mission.Client": "Equinor",
                "ImageMetadata.Timestamp": inspection.metadata.start_time.isoformat(),  # noqa: E501
                "ImageMetadata.X": str(inspection.metadata.pose.position.x),
                "ImageMetadata.Y": str(inspection.metadata.pose.position.y),
                "ImageMetadata.Y": str(inspection.metadata.pose.position.z),
                "ImageMetadata.CameraOrientation1": str(array_of_orientation[0]),
                "ImageMetadata.CameraOrientation2": str(array_of_orientation[1]),
                "ImageMetadata.CameraOrientation3": str(array_of_orientation[2]),
                "ImageMetadata.CameraOrientation4": str(array_of_orientation[3]),
                "ImageMetadata.AnalysisMethods": str(inspection.metadata.analysis),
                "ImageMetadata.Description": str(inspection.metadata.additional),
                "ImageMetadata.FunctionalLocation": inspection.metadata.tag_id  # noqa: E501
                if inspection.metadata.tag_id
                else "NA",
                "Filename": filename,
                "AttachedFile": (filename, inspection.data),
            }
        )
        return multiform_body

    @staticmethod
    def _construct_multiform_request_video(
        filename: str,
        inspection: Inspection,
        metadata: MissionMetadata,
    ):
        array_of_orientation = (
            inspection.metadata.pose.orientation.to_quat_array().tolist()
        )
        multiform_body: MultipartEncoder = MultipartEncoder(
            fields={
                "PlantFacilitySAPCode": metadata.plant_code,
                "InstCode": metadata.plant_short_name,
                "InternalClassification": metadata.data_classification,
                "IsoCountryCode": "NO",
                "Geodetic.CoordinateReferenceSystemCode": metadata.coordinate_reference_system,  # noqa: E501
                "Geodetic.VerticalCoordinateReferenceSystemCode": metadata.vertical_reference_system,  # noqa: E501
                "Geodetic.OrientationReferenceSystem": metadata.media_orientation_reference_system,  # noqa: E501
                "SensorCarrier.SensorCarrierId": metadata.robot_id,
                "SensorCarrier.ModelName": metadata.robot_model,
                "Mission.MissionId": metadata.mission_id,
                "Mission.Client": "Equinor",
                "VideoMetadata.Timestamp": inspection.metadata.start_time.isoformat(),  # noqa: E501
                "VideoMetadata.Duration": str(inspection.metadata.duration),  # type: ignore
                "VideoMetadata.X": str(inspection.metadata.pose.position.x),
                "VideoMetadata.Y": str(inspection.metadata.pose.position.y),
                "VideoMetadata.Y": str(inspection.metadata.pose.position.z),
                "VideoMetadata.CameraOrientation1": str(array_of_orientation[0]),
                "VideoMetadata.CameraOrientation2": str(array_of_orientation[1]),
                "VideoMetadata.CameraOrientation3": str(array_of_orientation[2]),
                "VideoMetadata.CameraOrientation4": str(array_of_orientation[3]),
                "VideoMetadata.AnalysisMethods": str(inspection.metadata.analysis),
                "VideoMetadata.Description": str(inspection.metadata.additional),
                "VideoMetadata.FunctionalLocation": inspection.metadata.tag_id  # noqa: E501
                if inspection.metadata.tag_id
                else "NA",
                "Filename": filename,
                "AttachedFile": (filename, inspection.data),
            }
        )
        return multiform_body
